


#include "psrdb.h"
#include <tostring.h>
#include <mysql++.h>
#include <Pulsar/Archive.h>
#include <Pulsar/ObsExtension.h>
#include <Pulsar/Integration.h>
#include <Pulsar/FITSHdrExtension.h>
#include <Pulsar/FITSSUBHdrExtension.h>
#include <Pulsar/Backend.h>
#include <Pulsar/WidebandCorrelator.h>
#include <Pulsar/Receiver.h>
#include <Pulsar/Telescope.h>
#include <dirutil.h>
#include <sstream>
#include <tempo++.h>



using mysqlpp::Query;
using mysqlpp::Result;
using mysqlpp::Row;
using std::cerr;
using std::endl;
using Pulsar::Archive;
using Pulsar::ObsExtension;
using Pulsar::FITSHdrExtension;
using Pulsar::FITSSUBHdrExtension;
using Pulsar::Backend;
using Pulsar::WidebandCorrelator;
using Pulsar::Receiver;
using Pulsar::Telescope;
using std::ostringstream;





/// Constructor

psrdb::psrdb()
{
}



/// Connect to the pulsararchive database using the parameters given,
/// If we failed to connect for any reason, simply throw an exception.


void psrdb::Connect( const char *db_name,
                     const char *db_hostname,
                     const char *db_username,
                     const char *db_password )
{
  if( !db.connect( db_name, db_hostname, db_username, db_password ) )
    throw( FailedConnection() );
}




/// Mark all the entries in the observations and cals tables as being
/// asbent. As we update entries we will mark the as present, and we
/// will know any that don't get marked as present can be deleted from
/// the database.

void psrdb::MarkObservations( void )
{
  Query obs_query = db.query();
  
  obs_query << "UPDATE observations SET present = 0 WHERE deleted = 0";
  obs_query.execute();
  
  Query cals_query = db.query();
  cals_query << "UPDATE cals SET present = 0 WHERE deleted = 0";
  cals_query.execute();
}




/// Purge any records in the observations table that haven't been marked as present.

void psrdb::PurgeMissingObservations()
{
  Query purge_query = db.query();
  
  purge_query << "UPDATE observations SET deleted = 1 WHERE present = 0 AND deleted = 0";
  
  purge_query.execute();
}



/// Purge any records in the cals table that haven't been marked as present

void psrdb::PurgeMissingCals()
{
  Query purge_query = db.query();
  
  purge_query << "UPDATE cals SET deleted = 1 WHERE present = 0 AND deleted = 0";
  
  purge_query.execute();
}


/// Calculate the galactic latitude and longitude

void psrdb::CalculateGalactic( double rajd, double decjd, double &gl, double &gb )
{
    double pi=asin(1) * 2.;
    double pion180 = pi/180.;
    double NGP_RA = 192.859508 * pion180;
    double NGP_DEC = 27.128336 * pion180;
    double ASC_NODE = 32.932;
    
    double draR = rajd*pion180 - NGP_RA;
    double decR = decjd*pion180;
    double sinb = cos(decR) * cos(NGP_DEC) * cos(draR) + sin(decR) * sin(NGP_DEC);
    gb = asin(sinb);/* # this is the latitude, but in radians.*/
    
    double sinl = (sin(decR) * cos(NGP_DEC) - cos(decR) * cos(draR) * sin(NGP_DEC)) / cos(gb);
    double cosl = cos(decR) * sin(draR) / cos(gb);
    
//     # Need to get the correct quadrant, as this isn't preserved by
//     # atan, which returns angle between -90 and 90.
    double gl_raw = atan(sinl/cosl);
   
    if(sinl > 0){
	if(cosl > 0 ){ gl = gl_raw; }
	else { gl = gl_raw + pi; }
    }
    else {
	if(cosl > 0){ gl = gl_raw + 2.*pi; }
	else{ gl = gl_raw + pi; }
    }
/*    # Now put them into degrees.*/
    gb = gb / pion180;
    gl = (gl / pion180) + ASC_NODE;
}

/// Read the archive parameters from the archive and put them in the ObservationDetails object.

void psrdb::GetArchiveDetails( ObservationDetails &details, Reference::To<Archive> arch )
{
  Reference::To<ObsExtension> obs_ext = arch->get<ObsExtension>();
  Reference::To<FITSHdrExtension> fits_ext = arch->get<FITSHdrExtension>();
  Reference::To<FITSSUBHdrExtension> fits_sub_ext = arch->get<FITSSUBHdrExtension>();
  Reference::To<Backend> backend = arch->get<Backend>();
  Reference::To<WidebandCorrelator> wbcorr = arch->get<WidebandCorrelator>();
  Reference::To<Receiver> receiver = arch->get<Receiver>();
  Reference::To<Telescope> telescope = arch->get<Telescope>();
  
  details.src_name = arch->get_source();
  if( obs_ext )
  {
    details.project_id = obs_ext->get_project_ID();
  }
  else
  {
    details.project_id = "";
  }
  details.raj = arch->get_coordinates().ra().getHMS();
  details.decj = arch->get_coordinates().dec().getDMS();
  if( backend )
  {
    details.data_type = backend->get_name();
    if( details.data_type.find("DFB") != details.data_type.npos )
      details.data_type = "DFB";
  }
  else if( wbcorr )
  {
    details.data_type = "WBC";
  }
  details.obsfreq = arch->get_centre_frequency();
  details.bw = arch->get_bandwidth();
  details.scanlen = arch->integration_length();

  MJD stt = arch->start_time();
  details.MJD = fromstring<double>( tostring( stt, 15 ) );
  char date_string[30];
  stt.datestr( date_string, 30, "%Y-%m-%d" );
  details.date = date_string;

  details.rajd = arch->get_coordinates().ra().getDegrees();
  details.decjd = arch->get_coordinates().dec().getDegrees();
  CalculateGalactic( details.rajd, details.decjd, details.gl, details.gb );

  if( !fits_ext )
  {
    details.bmaj = 0.0;
    details.bmin = 0.0;
    details.bpa = 0.0;
    details.ut = "UNDEF";
  }
  else
  {
    details.bmaj = fits_ext->get_bmaj();
    details.bmin = fits_ext->get_bmin();
    details.bpa = fits_ext->get_bpa();
    details.ut = fits_ext->get_stt_time();
    details.hdrver = fits_ext->get_hdrver();
  }

  details.dm = arch->get_dispersion_measure();
  details.period = arch->get_Integration(0)->get_folding_period();
  details.nbin = arch->get_nbin();
  details.npol = arch->get_npol();
  details.nchan = arch->get_nchan();
  details.nsub = arch->get_nsubint();
  if( fits_sub_ext )
  {
    details.tsamp = fits_sub_ext->get_tsamp();
    details.nbits = fits_sub_ext->get_nbits();
  }
  else
  {
    details.tsamp = 0.0;
    details.nbits = 0;
  }
  if( wbcorr )
  {
    details.cnfg = wbcorr->get_config();
    details.nbeam = wbcorr->get_nrcvr();
  }
//   details.inst = ;
  if( receiver )
    details.rcvr = receiver->get_name();

//   details.survery = ;
  
  if( obs_ext )
    details.telescope = obs_ext->get_telescope();
  else if( telescope )
    details.telescope = telescope->get_name();
  else
    details.telescope = "unknown";
  
  // if we didn't get a telescope, or get_name returne "unknown" then assume PARKES
  if( details.telescope == "unknown" )
    details.telescope = "PARKES";

  ostringstream s;
  s << Tempo::code(arch->get_telescope());
  details.site = s.str();

  if( obs_ext )
    details.observer = obs_ext->get_observer();
}


/// This function retreives the details of an observation given the data_loc/filename/prefix
/// It opens the file (or first file if multiple files) and calls GetArchiveDetails to read
/// the actual details.

void psrdb::GetObsDetails( const ObservationEntry &entry, ObservationDetails &details )
{
  if( entry.prefix == "\\0" )
  {
    string archive_path = entry.data_loc + entry.filename;
    Reference::To<Archive> arch = Archive::load( archive_path );
    if( !arch )
      throw( GenericException( (string("Failed to load archive") + archive_path).c_str() ) );
    
    GetArchiveDetails( details, arch );
    details.file_size_bytes = filesize( archive_path.c_str() );
  }
  else
  {
    Reference::To<Archive> first_arch = Archive::load( entry.files.front() );
    if( !first_arch )
      throw( GenericException( (string("Failed to load archive") + entry.files.front()).c_str() ) );

    GetArchiveDetails( details, first_arch );
    
    // Special case, if we have mutliple files, the number of subints is 
    // the number of files. ie always 1 subint per file.
    details.nsub = entry.files.size();
    
    unsigned long each_file_size = filesize( entry.files.front().c_str() );
    details.file_size_bytes = each_file_size * entry.files.size();
  }
}


/// This function takes an ObservationEntry that is either for an observation
/// or calibration and checks to see if an entry already exists in the database
/// for it. If not, it will create a new entry.

void psrdb::CheckObsCals( const ObservationEntry &entry, string table_name )
{
  cerr << entry.data_loc << entry.filename << entry.prefix << " ... ";
  
  try
  {
    Query q = db.query();
  
    q << " SELECT file_id ";
    q << " FROM " << table_name << " ";
    q << " WHERE data_loc = \"" << entry.data_loc << "\"";
    q << " AND filename = \"" << entry.filename << "\"";
    q << " AND prefix = '" << entry.prefix << "' ";
    q << " AND deleted = 0 ";

    Result res = q.store();
  
    if( res.num_rows() != 0 )
    {
      Row row = res.fetch_row();
      string file_id = row["file_id"].get_string();

      Query present_query = db.query();
      present_query << " UPDATE " << table_name << " SET present = 1 WHERE file_id = " << file_id;
      present_query.execute();
      cerr << "UPDATED" << endl;
    }
    else
    {
      ObservationDetails details;
      GetObsDetails( entry, details );
    
      Query insert_query = db.query();
      insert_query << " INSERT into " << table_name << "(data_loc,filename,prefix,obsfreq,bw,scanlen,nbin,nchan,npol,nsub,raj,decj,rajd,decjd,mjd,bmaj,bmin,bpa,dm,period,tsamp,nbits,nbeam,cnfg,rcvr,hdrver,survey,telescope,site,obsrvr,date,ut,src_name,project_id,data_type,file_size_bytes,gl,gb,present) ";
      insert_query << " VALUES( \"" << entry.data_loc << "\", ";
      insert_query << " \"" << entry.filename << "\", ";
      insert_query << " '" << entry.prefix << "', ";
      insert_query << details.obsfreq << ", ";
      insert_query << details.bw << ", ";
      insert_query << details.scanlen << ", ";
      insert_query << details.nbin << ", ";
      insert_query << details.nchan << ", ";
      insert_query << details.npol << ", ";
      insert_query << details.nsub << ", ";
      insert_query << "'" << details.raj << "', ";
      insert_query << "'" << details.decj << "', ";
      insert_query << details.rajd << ", ";
      insert_query << details.decjd << ", ";
      insert_query << details.MJD << ", ";
      insert_query << details.bmaj << ", ";
      insert_query << details.bmin << ", ";
      insert_query << details.bpa << ", ";
      insert_query << details.dm << ", ";
      insert_query << details.period << ", ";
      insert_query << details.tsamp << ", ";
      insert_query << details.nbits << ", ";
      insert_query << details.nbeam << ", ";
      insert_query << "'" << details.cnfg << "', ";
      insert_query << "'" << details.rcvr << "', ";
      insert_query << "'" << details.hdrver << "', ";
      insert_query << "'" << details.survey << "', ";
      insert_query << "'" << details.telescope << "', ";
      insert_query << "'" << details.site << "', ";
      insert_query << "'" << details.observer << "', ";
      insert_query << "'" << details.date << "', ";
      insert_query << "'" << details.ut << "', ";
      insert_query << "'" << details.src_name << "', ";
      insert_query << "'" << details.project_id << "', ";
      insert_query << "'" << details.data_type << "', ";
      insert_query << details.file_size_bytes << ", ";
      insert_query << details.gl << ", ";
      insert_query << details.gb << ", ";
      insert_query << " 1 )"; // present = 1
    
      try
      {
        insert_query.execute();
	cerr << "DONE" << endl;
      }
      catch( Exception e )
      {
	cerr << "FAILED" << endl;
	cerr << "---------------------------------------" << endl;
	cerr << "insert query failed: " << e.what() << endl;
	cerr << "full query was: " << insert_query.preview() << endl;
	cerr << "---------------------------------------" << endl;
      }
    }
  }
  catch( Error e )
  {
//     cerr << "received an exception from psrchive on file " << entry.data_loc << entry.filename << endl;
    cerr << "FAILED" << endl;
  }
}


